@@ -28,12 +28,12 @@ class SimpleWorker(BaseWorker):
     """Simple one-thread worker"""
 
     def run(self) -> None:
-        self._thread = threading.Thread(target=self._run_thread)  # pylint: disable=attribute-defined-outside-init
+        self._thread = threading.Thread(target=self._run_thread)
         self._thread.daemon = True
         self._thread.start()
 
     def _run_thread(self) -> None:
-        logger.info('[SimpleWorker] started')
+        logger.info("[SimpleWorker] started")
 
         while self._is_enabled:
             try:
